"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.orderTypes = void 0;
const LRUCache = require("lru-cache");
const expressions_1 = require("../expressions");
const TermTransformer_1 = require("../transformers/TermTransformer");
const Consts_1 = require("./Consts");
const DateTimeHelpers_1 = require("./DateTimeHelpers");
const Err = require("./Errors");
const TypeHandling_1 = require("./TypeHandling");
// Determine the relative numerical order of the two given terms.
/**
 * @param enableExtendedXSDTypes System will behave like when this was true. @deprecated
 */
function orderTypes(termA, termB, strict = false, typeDiscoveryCallback, typeCache, enableExtendedXSDTypes) {
    // Check if terms are the same by reference
    if (termA === termB) {
        return 0;
    }
    // We handle undefined that is lower than everything else.
    if (termA === undefined) {
        return -1;
    }
    if (termB === undefined) {
        return 1;
    }
    //
    if (termA.termType !== termB.termType) {
        return _TERM_ORDERING_PRIORITY[termA.termType] < _TERM_ORDERING_PRIORITY[termB.termType] ? -1 : 1;
    }
    // Check exact term equality
    if (termA.equals(termB)) {
        return 0;
    }
    // Handle quoted triples
    if (termA.termType === 'Quad' && termB.termType === 'Quad') {
        const orderSubject = orderTypes(termA.subject, termB.subject, strict, typeDiscoveryCallback, typeCache, enableExtendedXSDTypes);
        if (orderSubject !== 0) {
            return orderSubject;
        }
        const orderPredicate = orderTypes(termA.predicate, termB.predicate, strict, typeDiscoveryCallback, typeCache, enableExtendedXSDTypes);
        if (orderPredicate !== 0) {
            return orderPredicate;
        }
        const orderObject = orderTypes(termA.object, termB.object, strict, typeDiscoveryCallback, typeCache, enableExtendedXSDTypes);
        if (orderObject !== 0) {
            return orderObject;
        }
        return orderTypes(termA.graph, termB.graph, strict, typeDiscoveryCallback, typeCache, enableExtendedXSDTypes);
    }
    // Handle literals
    if (termA.termType === 'Literal') {
        return orderLiteralTypes(termA, termB, typeDiscoveryCallback, typeCache);
    }
    // Handle all other types
    if (strict) {
        throw new Err.InvalidCompareArgumentTypes(termA, termB);
    }
    return comparePrimitives(termA.value, termB.value);
}
exports.orderTypes = orderTypes;
function orderLiteralTypes(litA, litB, typeDiscoveryCallback, typeCache) {
    const defaultTimezone = { zoneHours: 0, zoneMinutes: 0 };
    const openWorldType = {
        discoverer: typeDiscoveryCallback || (() => 'term'),
        cache: typeCache || new LRUCache(),
    };
    const termTransformer = new TermTransformer_1.TermTransformer(openWorldType);
    const myLitA = termTransformer.transformLiteral(litA);
    const myLitB = termTransformer.transformLiteral(litB);
    const typeA = myLitA.dataType;
    const typeB = myLitB.dataType;
    const superTypeDictA = (0, TypeHandling_1.getSuperTypeDict)(typeA, openWorldType);
    const superTypeDictB = (0, TypeHandling_1.getSuperTypeDict)(typeB, openWorldType);
    // Special handling of specific datatypes
    if (!(0, expressions_1.isNonLexicalLiteral)(myLitA) && !(0, expressions_1.isNonLexicalLiteral)(myLitB)) {
        if (Consts_1.TypeURL.XSD_BOOLEAN in superTypeDictA && Consts_1.TypeURL.XSD_BOOLEAN in superTypeDictB ||
            Consts_1.TypeAlias.SPARQL_NUMERIC in superTypeDictA && Consts_1.TypeAlias.SPARQL_NUMERIC in superTypeDictB ||
            Consts_1.TypeURL.XSD_STRING in superTypeDictA && Consts_1.TypeURL.XSD_STRING in superTypeDictB) {
            return comparePrimitives(myLitA.typedValue, myLitB.typedValue);
        }
        if (Consts_1.TypeURL.XSD_DATE_TIME in superTypeDictA && Consts_1.TypeURL.XSD_DATE_TIME in superTypeDictB) {
            return comparePrimitives((0, DateTimeHelpers_1.toUTCDate)(myLitA.typedValue, defaultTimezone).getTime(), (0, DateTimeHelpers_1.toUTCDate)(myLitB.typedValue, defaultTimezone).getTime());
        }
        if (Consts_1.TypeURL.RDF_LANG_STRING in superTypeDictA && Consts_1.TypeURL.RDF_LANG_STRING in superTypeDictB) {
            const compareType = comparePrimitives(myLitA.typedValue, myLitB.typedValue);
            if (compareType !== 0) {
                return compareType;
            }
            return comparePrimitives(myLitA.language, myLitB.language);
        }
    }
    // Fallback to string-based comparison
    const compareType = comparePrimitives(typeA, typeB);
    if (compareType !== 0) {
        return compareType;
    }
    return comparePrimitives(myLitA.str(), myLitB.str());
}
function comparePrimitives(valueA, valueB) {
    // eslint-disable-next-line @typescript-eslint/no-extra-parens
    return valueA === valueB ? 0 : (valueA < valueB ? -1 : 1);
}
// SPARQL specifies that blankNode < namedNode < literal.
const _TERM_ORDERING_PRIORITY = {
    Variable: 0,
    BlankNode: 1,
    NamedNode: 2,
    Literal: 3,
    Quad: 4,
    DefaultGraph: 5,
};
//# sourceMappingURL=Ordering.js.map